/*
 * Created on Jul 6, 2005
 *
 */
package tacle.core;

import org.eclipse.jdt.core.IMethod;
import java.util.Set;

/**
 * Tacle-specific representation of a call site. Client code should
 * not implement this interface: instances should be obtained through
 * methods <code>getXYZ</code> from <code>CallGraphAnalysis</code>.
 * <p> Eventually we may add subinterfaces <code>StaticCallSite</code>
 * and possibly <code>ConstructorCallSite</code>. But for now, just
 * use one subinterface <code>InstanceCallSite</code>. 
 */

public interface ICallSite {

    /** 
     * Returns a reference to the IMethod for the enclosing
     * method. Could be <code>null</code> due to the limitations of
     * <code>IType</code> and <code>IMethod</code>. 
     */
    IMethod getMethod(); 
    
    /** 
     * Start position in the source code, copied from the
     * corresponding AST node. If the method does not come from source
     * code (e.g., it is a method from a .class file), returns -1.
     */
    int getStartPosition();

    /** 
     * Length of the source segment, copied from the corresponding AST
     * node. If the method does not come from source code (e.g., it is
     * a method from a .class file), returns -1.
     */
    int getLength();

    /** 
     * Returns a set of <code>IMethod</code> references for all
     * methods called by the call site. Not guarantted to be complete,
     * due to the limitations of <code>IType</code> and
     * <code>IMethod</code>. 
     * 
     * The current version does not deal with reflective calls and throw 
     * clauses. Methods declared in anonymous classes are not reported.
     * The body of a method declared in an anonymous class created that
     * is declared in a method may be wrongly reported as belonging
     * to the enclosing method. See 
     * http://presto.cse.ohio-state.edu/tacle/index.html
     * for more details on the current implementation.   
     * 
     * Current version works only with features of Java 1.4.
     */    
    Set getCalledIMethods();

}
    
