package tacle.example.popups.actions;

import org.eclipse.jface.action.IAction;
import org.eclipse.jface.dialogs.MessageDialog;
import org.eclipse.jface.viewers.ISelection;
import org.eclipse.swt.widgets.Shell;
import org.eclipse.ui.IObjectActionDelegate;
import org.eclipse.ui.IWorkbenchPart;
import tacle.examples.simple.*;
import org.eclipse.jdt.core.IMethod;
import org.eclipse.jface.viewers.IStructuredSelection;
import org.eclipse.core.runtime.jobs.Job;
import org.eclipse.core.runtime.jobs.ILock; 
import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.Platform;
import org.eclipse.core.runtime.Status;
import org.eclipse.core.runtime.IStatus;
import org.eclipse.ui.progress.IProgressConstants;
import org.eclipse.jface.action.Action;

//This is an example of how to run the TACLE API as a job


public class HelloTacleActionJob implements IObjectActionDelegate {

	private ISelection selection;
	/**
	 * Constructor for Action1.
	 */
	public HelloTacleActionJob() {
		super();
	}

	/**
	 * @see IObjectActionDelegate#setActivePart(IAction, IWorkbenchPart)
	 */
	public void setActivePart(IAction action, IWorkbenchPart targetPart) {
	}

	/**
	 * @see IActionDelegate#run(IAction)
	 */
	public void run(IAction action) {

		final int[] x = new int[1];

		Job j = new Job("TacleAnalysis") {
			protected IStatus run(IProgressMonitor monitor) {
				Object element = getSingleElement(selection);
				IMethod method = getSelectedMethod(element);
				ILock accessLock = Platform.getJobManager().newLock();
				try {
					accessLock.acquire();
					HelloTacle ht = new HelloTacle();
					x[0] = ht.run(method);
				} finally {
					accessLock.release();
				}
				accessLock.release();
				return new Status(IStatus.INFO,
						"edu.ohio-state.cse.presto.tacle", IStatus.OK,
						"There are " + x[0] + " reachable methods.", null);
			}
		};

		System.out.println("Job created");
		IAction jobAction = new Action() {
			public void run() {
				Shell shell = new Shell();
				MessageDialog.openInformation(shell, "Tacle Plug-in",
						"There are " + x[0] + " reachable methods");
			}
		};
		j.setProperty(IProgressConstants.KEEP_PROPERTY, Boolean.TRUE);
		j.setProperty(IProgressConstants.ACTION_PROPERTY, jobAction);
		j.schedule();
		System.out.println("Job scheduled");
	}



	
/*	public void run(IAction action) {
		Shell shell = new Shell();
		Object element = getSingleElement(selection);

		IMethod method = getSelectedMethod(element);
		
		HelloTacle ht =new HelloTacle();
		MessageDialog.openInformation(
			shell,
			"Tacle Plug-in",
			"There are "+ht.run(method)+" reachable methods");
	}*/

	/**
	 * @see IActionDelegate#selectionChanged(IAction, ISelection)
	 */
	public void selectionChanged(IAction action, ISelection selection) {
		this.selection = selection;
	}

	public static Object getSingleElement(ISelection s) {
		if (!(s instanceof IStructuredSelection))
			return null;
		IStructuredSelection selection = (IStructuredSelection) s;
		if (selection.size() != 1)
			return null;
		return selection.getFirstElement();

	}


	private IMethod getSelectedMethod(Object element) {
		IMethod method = null;
		if (element instanceof IMethod) {
			method = (IMethod) element;
		}
		if(method.getCompilationUnit() == null){
			throw new RuntimeException("The selected method, "+method.getElementName()+", doesn't have a compilation unit");		
		}
		return method;
	}
}
