/*
 * Created on May 19, 2005
 *
 */
package tacle.internal.rta;

import org.eclipse.jdt.core.dom.ASTNode;
import org.eclipse.jdt.core.dom.ASTVisitor;
import org.eclipse.jdt.core.dom.AnonymousClassDeclaration;
import org.eclipse.jdt.core.dom.ClassInstanceCreation;
import org.eclipse.jdt.core.dom.ConstructorInvocation;
import org.eclipse.jdt.core.dom.IBinding;
import org.eclipse.jdt.core.dom.IMethodBinding;
import org.eclipse.jdt.core.dom.ITypeBinding;
import org.eclipse.jdt.core.dom.IVariableBinding;
import org.eclipse.jdt.core.dom.MethodDeclaration;
import org.eclipse.jdt.core.dom.MethodInvocation;
import org.eclipse.jdt.core.dom.Modifier;
import org.eclipse.jdt.core.dom.SimpleName;
import org.eclipse.jdt.core.dom.TypeDeclaration;
//import org.eclipse.jdt.core.dom.StringLiteral;
import org.eclipse.jdt.core.dom.SuperConstructorInvocation;
import org.eclipse.jdt.core.dom.SuperMethodInvocation;
//import org.eclipse.jdt.core.dom.TypeDeclaration;
import org.eclipse.jdt.internal.corext.util.JdtFlags;

import tacle.internal.rta.data.ASTType;
import tacle.internal.rta.data.IWorklistType;
import tacle.internal.rta.data.IWorklistMethod;
import tacle.internal.rta.data.InfoDictionary;

/**
 * @author Mariana L Barca
 * 
 * visits AST of a method not to be used for the whole AST of a compilation unit
 * in case of a special method for that case, subclass this class and use a
 * filter
 */
public class MethodVisitor extends ASTVisitor {

	protected RTAMethodWorklist worklist;

	protected IWorklistMethod method;

	private MethodVisitor() {
	}

	MethodVisitor(RTAMethodWorklist worklist, IWorklistMethod method) {
		this.worklist = worklist;
		this.method = method;
	}

	//////////// visit methods /////////////
	public boolean visit(ClassInstanceCreation node) {
		//		System.out.println(" CLASS INSTANCE CREATION:" + node);
		ITypeBinding tb = node.resolveTypeBinding();
		
//		// test
//		if (tb == null)
//			System.out.println("MethodVisitor:ClassINstanceCreation");
		
		if (tb.isAnonymous()) {
			// add inner type in the dictionary for future reference
			IWorklistType innerCls = InfoDictionary.v().getTypeForBinding(tb);
			if (innerCls instanceof ASTType) {
				// anon type decl starts after "new ", hence + pos (pos = 4
				// usually)
				int pos = node.getStartPosition()
						+ getAnonClassPosition(node.toString());
//				System.out.println("CLASS INSTANCE CREATION AT "
//						+ node.getParent().getClass());
				((ASTType) innerCls).setPosition(pos, getTopLevelParent(node));
			}
		}
		worklist.addToTypeHierarchy(tb, true);
		worklist.addInstantiatedClass(tb);
		// constructor not called explicitly
		IMethodBinding mb = node.resolveConstructorBinding();
		
//		// test
//		if (mb == null)
//			System.out.println("MethodVisitor:ClassINstanceCreation 2");

		if (mb != null) {
			addConstructor(mb, node.getStartPosition(), node.getLength());
		}
		return true;
	}

	public boolean visit(ConstructorInvocation node) {
		//System.out.println(method + " CONSTRUCTOR INVOCATION:" + node);
		IMethodBinding mb = node.resolveConstructorBinding();
		
//		// test
//		if (mb == null)
//			System.out.println("MethodVisitor:ConstructorInvocation");
		
		if (mb != null) {
			addConstructor(mb, node.getStartPosition(), node.getLength());
		}
		return true;
	}

	public boolean visit(MethodInvocation node) {
		//System.out.println("METHOD INVOCATION:" + node);
		IMethodBinding mb = node.resolveMethodBinding();
		
//		// test
//		if (mb == null)
//			System.out.println("MethodVisitor:MethodInvocation");
		
		if (mb != null) {
			int modifiers = mb.getModifiers();
			if (Modifier.isStatic(modifiers)) {
				//System.out.println("Static");
				addNonVirtualMethod(mb, node.getStartPosition(), node
						.getLength());
			} else {
				//System.out.println("Virtual");
				addVirtualMethod(mb, node.getStartPosition(), node.getLength());
			}
		}
		return true;
	}

	// this can be the name of a static field
	public boolean visit(SimpleName node) {
		IBinding binding = node.resolveBinding();
		
//		// test
//		if (binding == null)
//			System.out.println("MethodVisitor:SimpleName");

		if (binding != null && binding.getKind() == IBinding.VARIABLE) {
			IVariableBinding iVarBinding = (IVariableBinding) binding;
			if (iVarBinding.isField()) {
				if (JdtFlags.isStatic(iVarBinding)) {
					//System.out.println("SIMPLE NAME, STATIC FIELD:" + node);
					if (iVarBinding.getType().isClass())
						worklist
								.addToTypeHierarchy(iVarBinding.getType(), true);
				}
			}
		}
		return true;
	}

	// invocation through super are not virtual calls
	public boolean visit(SuperConstructorInvocation node) {
		//System.out.println("SUPER CONSTRUCTOR INVOCATION:" + node);
		IMethodBinding mb = node.resolveConstructorBinding();
		
//		// test
//		if (mb == null)
//			System.out.println("MethodVisitor:SuperConstructorInvocation");
		
		if (mb != null) {
			addConstructor(mb, node.getStartPosition(), node.getLength());
		}
		return true;
	}

	// invocation through super are not virtual calls
	public boolean visit(SuperMethodInvocation node) {
		//System.out.println("SUPER METHOD INVOCATION:" + node);
		IMethodBinding mb = node.resolveMethodBinding();
		
//		// test
//		if (mb == null)
//			System.out.println("MethodVisitor:SuperMethodInvocation");
		
		if (mb != null) {
			addNonVirtualMethod(mb, node.getStartPosition(), node.getLength());
		}
		return true;
	}

	// types declared inside the method are not visited
	public boolean visit(TypeDeclaration node) {
		ITypeBinding itb = node.resolveBinding();
		
//		// test
//		if (itb == null)
//			System.out.println("MethodVisitor:TypeDeclaration");
		
		IWorklistType temp = InfoDictionary.v().getTypeForBinding(itb);
		if (temp instanceof ASTType) {
			//System.out.println("---" + ((ASTType)temp).getCompleteName());
			((ASTType) temp).setPosition(node.getStartPosition(), getTopLevelParent(node));
		}
		return false;
	}

	///// end visit methods

	///// private helpers

	// checks if class represented by itb implements
	// the interface interf directly or indirectly
	private boolean implementsInterface(ITypeBinding itb, String interf) {
		if (itb.getQualifiedName().equals(interf))
			return true;
		ITypeBinding[] interfs = itb.getInterfaces();
		for (int i = 0; i < interfs.length; i++) {
			if (implementsInterface(interfs[i], interf))
				return true;
		}
		return false;
	}

	// not sure this is necessary
	private ITypeBinding getTopLevel(ITypeBinding itb) {
		ITypeBinding result = itb;
		while (!result.isTopLevel()) {
			result = result.getDeclaringClass();
		}
		return result;
	}

	private void addConstructor(IMethodBinding mb, int pos, int len) {
		IWorklistType[] paramTypes = getParameterTypes(mb);
		IWorklistType declClass = InfoDictionary.v().getTypeForBinding(
				mb.getDeclaringClass());
		IWorklistType top = declClass;
		ITypeBinding topLevel = getTopLevel(mb.getDeclaringClass());
		if (topLevel != mb.getDeclaringClass())
			top = InfoDictionary.v().getTypeForBinding(topLevel);
		worklist.addCGEdge(method, null, pos, len, mb.getKey(), "<init>",
				declClass, top, InfoDictionary.v().getOtherTypeForName("void"),
				paramTypes, false);
		// add declaring class to type hierarchy
		worklist.addToTypeHierarchy(mb.getDeclaringClass(), true);
	}

	private void addNonVirtualMethod(IMethodBinding mb, int pos, int len) {
		IWorklistType[] paramTypes = getParameterTypes(mb);
		IWorklistType declClass = InfoDictionary.v().getTypeForBinding(
				mb.getDeclaringClass());
		IWorklistType top = declClass;
		ITypeBinding topLevel = getTopLevel(mb.getDeclaringClass());
		if (topLevel != mb.getDeclaringClass())
			top = InfoDictionary.v().getTypeForBinding(topLevel);
		IWorklistType returnType = InfoDictionary.v().getTypeForBinding(
				mb.getReturnType());
		worklist.addCGEdge(method, null, pos, len, mb.getKey(), mb.getName(),
				declClass, top, returnType, paramTypes, false);
		// add declaring class to type hierarchy
		worklist.addToTypeHierarchy(mb.getDeclaringClass(), true);
	}

	private void addVirtualMethod(IMethodBinding mb, int pos, int len) {
		ITypeBinding tb = mb.getDeclaringClass();
		// add declaring class to the hierarchy but not initialized yet
		worklist.addToTypeHierarchy(tb, false);
		IWorklistType[] pTypes = getParameterTypes(mb);
		IWorklistType declCls = InfoDictionary.v().getTypeForBinding(tb);
		if (mb.getName().equals("start") && pTypes.length == 0
				&& mb.getReturnType().getQualifiedName().equals("void")
				&& implementsInterface(tb, "java.lang.Runnable")) {
			// this is thread start, call run
			worklist.processVirtualCall(method, declCls, "run", pTypes, pos,
					len);
		} else if (mb.getName().equals("doPrivileged")
				&& pTypes.length >= 1
				&& mb.getDeclaringClass().getQualifiedName().equals(
					"java.security.AccessController")) {
			//System.out.println(" --- " + mb.getKey());
			worklist.processVirtualCall(method, pTypes[0], "run",
					new IWorklistType[0], -1, -1);
		} else {
			worklist.processVirtualCall(method, declCls, mb.getName(), pTypes,
					pos, len);
		}
	}

	private IWorklistType[] getParameterTypes(IMethodBinding mb) {
		ITypeBinding[] params = mb.getParameterTypes();
		IWorklistType[] pTypes = new IWorklistType[params.length];
		for (int i = 0; i < params.length; i++) {
			pTypes[i] = InfoDictionary.v().getTypeForBinding(params[i]);
		}
		return pTypes;
	}

	private int getAnonClassPosition(String anonDecl) {
		int pos = anonDecl.indexOf("new");
		pos += 4; // go over "new "
		boolean found = false;
		while (pos < anonDecl.length() && !found) {
			char c = anonDecl.charAt(pos);
			if (('A' <= c && c <= 'Z') || ('a' <= c && c <= 'z')) {
				found = true;
			} else {
				pos++;
			}
		}
		return pos;
	}

	private String getTopLevelParent(ASTNode node) {
		ASTNode result = node;
		while (!(result instanceof TypeDeclaration)
				|| !((TypeDeclaration) result).resolveBinding().isTopLevel()) {
			result = result.getParent();
		}
		return ((TypeDeclaration) result).resolveBinding().getQualifiedName();
	}
}
