/*
 * Created on May 19, 2005
 *
 */
package tacle.internal.rta;

import java.util.Set;
import java.util.HashSet;
import java.util.Map;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Vector;

import tacle.core.ICallGraphAnalysis;
import tacle.core.ICallSite;
import tacle.internal.types.TacleTypesException;
import tacle.internal.rta.data.ASTType;
import tacle.internal.rta.data.CallSite;
import tacle.internal.rta.data.InstanceCallSite;
import tacle.internal.rta.data.IWorklistMethod;
import tacle.internal.rta.data.IWorklistType;
import tacle.internal.rta.data.ASTMethod;

import org.eclipse.jdt.core.ICompilationUnit;
import org.eclipse.jdt.core.IField;
import org.eclipse.jdt.core.IInitializer;
import org.eclipse.jdt.core.IMethod;
import org.eclipse.jdt.core.IPackageFragment;
import org.eclipse.jdt.core.JavaModelException;
import org.eclipse.jdt.core.IJavaProject;
import org.eclipse.jdt.core.IJavaElement;
import org.eclipse.jdt.core.IType;
import org.eclipse.jdt.core.Signature;

/**
 * @author Mariana L Barca
 * 
 * singleton the main entry point to RTA
 */
public class RTAnalysis implements ICallGraphAnalysis {

	private Set reachableMethods; // member IWorklistMethod

	private Set initializedTypes; // member IWorklistClass

	private TypeManager typeMgr;

	private Set javaProjects;

	// map tp hold equivalent representations of methods
	// IMethod -> IWorklistMethod
	private Map externalToInternalMethods;

	// map tp hold equivalent representations of methods
	// IWorklistMethod -> IMethod
	private Map internalToExternalMethods;

	//	private Map externalToInternalTypes;
	//
	//	private Map internalToExternalTypes;

	private boolean methodsPopulated;

	// constructor
	public RTAnalysis() {
		reachableMethods = new HashSet();
		initializedTypes = new HashSet();
		javaProjects = new HashSet();

		externalToInternalMethods = new HashMap();
		internalToExternalMethods = new HashMap();
		//		externalToInternalTypes = new HashMap();
		//		internalToExternalTypes = new HashMap();

		methodsPopulated = false;

		addReachableMethods = new Vector();
		addInstantiatedClasses = new Vector();
	}

	private Vector addReachableMethods;

	public void addLibraryReachableMethod(String[] method) {
		addReachableMethods.add(method);
	}

	public void addReachableMethod(IMethod method) {
		addReachableMethods.add(method);
	}

	private Vector addInstantiatedClasses;

	public void addLibraryInstantiatedClass(String packageName, String className) {
		addInstantiatedClasses.add(new String[] { packageName, className });
	}

	public void addInstantiatedClass(IType type) {
		addInstantiatedClasses.add(type);
	}

	/**
	 * main method is the input of this analysis
	 */
	public void run(IMethod m) {
		// check if m is a main method
		//		try {
		//			if (!m.isMainMethod()) {
		//				return;
		//				//throw new RuntimeException(m + " is not a main method");
		//			}
		//		} catch (JavaModelException jme) {
		//			jme.printStackTrace();
		//			//throw new TacleTypesException(jme.getMessage());
		//		}

		// AST BASED ANALYSIS
		try {
			// mem and time measurements
			System.gc();
			System.gc();
			System.gc();
			System.gc();
			System.gc();
			long startMem = Runtime.getRuntime().totalMemory()
					- Runtime.getRuntime().freeMemory();
			long startTime = System.currentTimeMillis();

			// the actual execution starts here
			typeMgr = new TypeManager(m);
			RTAMethodWorklist wklist = new RTAMethodWorklist(typeMgr,
					reachableMethods, initializedTypes, javaProjects);

			// insert any additional inst classes or reachable methods
			// added by user
			for (Iterator it = addInstantiatedClasses.iterator(); it.hasNext();) {
				Object obj = it.next();
				if (obj instanceof IType) {
					IType ic = (IType) obj;
					String simpleName = ic.getElementName();
					String completeName = ic.getFullyQualifiedName();
					String packageName = ic.getPackageFragment()
							.getElementName();
					// ???
					String parentName = ic.getCompilationUnit()
							.getElementName();
					try {
						int position = ic.getSourceRange().getOffset();
						wklist.addUserStartupClass(packageName, completeName,
								simpleName, parentName, position);
					} catch (JavaModelException jme) {
						System.out.println(jme.getMessage());
					}
				} else if (obj instanceof String[]) {
					String[] ic = (String[]) obj;
					wklist.addLibraryStartupClass(ic);
				}
			}
			for (Iterator it = addReachableMethods.iterator(); it.hasNext();) {
				Object obj = it.next();
				if (obj instanceof IMethod) {
					IMethod rm = (IMethod) obj;
					String name = rm.getElementName();
					IType parent = rm.getDeclaringType();
					String parentPackageName = parent.getPackageFragment()
							.getElementName();
					String parentCompleteClassName = parent
							.getFullyQualifiedName();
					String parentSimpleClassName = parent.getElementName();
					try {
						if (rm.isConstructor())
							name = "<init>";
						int location = rm.getSourceRange().getOffset();
						wklist.addUserStartupMethod(name, parentPackageName,
								parentCompleteClassName, parentSimpleClassName,
								location);
					} catch (JavaModelException jme) {
						System.out.println(jme.getMessage());
					}
				} else if (obj instanceof String[]) {
					String[] rm = (String[]) obj;
					wklist.addLibraryStartupMethod(rm);
				}
			}

			wklist.buildMainMethod();
			wklist.build();

			// mem and time measurements
			System.out.println((System.currentTimeMillis() - startTime)
					/ 1000.0 + " seconds");
			System.gc();
			System.gc();
			System.gc();
			System.gc();
			System.gc();
			System.out.println((Runtime.getRuntime().totalMemory()
					- Runtime.getRuntime().freeMemory() - startMem)
					/ (1024 * 1024.0) + " MB");
			wklist.printStat();
		} catch (TacleTypesException tte) {
			tte.printStackTrace();
		}

		//Set set = getInitializedITypes();

	}

	// dummy implementations for interface methods
	// temporary fix to work with the examples
	// members of reachableMethods have a different type than expected
	public Set getReachableIMethods() {
		if (!methodsPopulated) {
			populateMethods();
		}
		return externalToInternalMethods.keySet();
		//return reachableMethods;
	}

	// dummy implementations of the interface methods
	public Set getInitializedITypes() {
		Set initializedITypes = new HashSet();
		for (Iterator it = initializedTypes.iterator(); it.hasNext();) {
			IWorklistType iwt = (IWorklistType) it.next();
			IType iType = getIType(iwt);
			initializedITypes.add(iType);
		}

		//        for(Iterator i = initializedITypes.iterator(); i.hasNext();){
		//            IType it = (IType)i.next();
		//            try{
		//                if(it == null){
		//                    System.out.println("WHY AM I GETTING A NULL ITYPE");
		//                }else if(it.isAnonymous() && !it.isBinary()) {
		//                    System.out.println(it.getFullyQualifiedName() + " " +it.gt);
		//                }
		//            }catch(Exception e){e.printStackTrace();}
		//        }

		return initializedITypes;
	}

	public Set getCalledIMethods(IMethod method) {
		if (!methodsPopulated) {
			populateMethods();
		}
		IWorklistMethod internalM = (IWorklistMethod) externalToInternalMethods
				.get(method);
		if (internalM == null) {
			return null;
		}
		Set set = internalM.getResolvedCallSites();
		Set result = new HashSet();
		for (Iterator it = set.iterator(); it.hasNext();) {
			ResolvedCallSite rcs = (ResolvedCallSite) it.next();
			Set targets = rcs.resolvedTargets;
			for (Iterator it1 = targets.iterator(); it1.hasNext();) {
				IWorklistMethod iwm = (IWorklistMethod) it1.next();
				IMethod tgt = (IMethod) internalToExternalMethods.get(iwm);
				if (tgt != null) {
					//System.out.println("ADDING: " + iwm + " " + tgt);
					result.add(tgt);
				}
			}
		}
		return result;
	}

	public Set getCallSites(IMethod method) {
		if (!methodsPopulated) {
			populateMethods();
		}
		IWorklistMethod internalM = (IWorklistMethod) externalToInternalMethods
				.get(method);
		if (internalM == null) {
			return null;
		}
		Set result = new HashSet();
		String internalMKey = internalM.getKey();
		Set resolvedSites = internalM.getResolvedCallSites();
		for (Iterator it = resolvedSites.iterator(); it.hasNext();) {
			ResolvedCallSite rcs = (ResolvedCallSite) it.next();
			Set targetMethods = new HashSet();
			Set targetMethodKeys = new HashSet();
			CallSite cs = null;
			if (rcs.isVirtual()) {
				cs = new InstanceCallSite(method, internalMKey, rcs.getPos(),
						rcs.getLength(), targetMethods, targetMethodKeys);
			} else {
				cs = new CallSite(method, internalMKey, rcs.getPos(), rcs
						.getLength(), targetMethods, targetMethodKeys);
			}
			Set targets = rcs.resolvedTargets;
			for (Iterator it1 = targets.iterator(); it1.hasNext();) {
				IWorklistMethod tgt = (IWorklistMethod) it1.next();
				IMethod iMtgt = (IMethod) internalToExternalMethods.get(tgt);
				if (iMtgt == null) {
					System.out.println("IMethod not found for " + tgt);
				} else {
					String tgtKey = tgt.getKey();
					targetMethods.add(iMtgt);
					targetMethodKeys.add(tgtKey);
				}
			}
			result.add(cs);
		}
		return result;
	}

	public boolean analyzedFromSourceCode(IMethod method) {
		if (!methodsPopulated) {
			populateMethods();
		}
		IWorklistMethod internalM = (IWorklistMethod) externalToInternalMethods
				.get(method);
		if (internalM == null) {
			return false;
		}
		//System.out.println("CLASS:" + internalM.getClass());
		return (internalM instanceof ASTMethod);
	}

	public ICallSite getCallSite(IMethod method, int pos, int length) {
		if (!methodsPopulated) {
			populateMethods();
		}
		IWorklistMethod internalM = (IWorklistMethod) externalToInternalMethods
				.get(method);
		if (internalM == null) {
			return null;
		}
		String internalMKey = internalM.getKey();
		Set resolvedSites = internalM.getResolvedCallSites();
		Iterator it = resolvedSites.iterator();
		while (it.hasNext()) {
			ResolvedCallSite rcs = (ResolvedCallSite) it.next();
			if (rcs.getPos() == pos && rcs.getLength() == length) {
				Set targetMethods = new HashSet();
				Set targetMethodKeys = new HashSet();
				CallSite cs = null;
				if (rcs.isVirtual()) {
					cs = new InstanceCallSite(method, internalMKey, pos,
							length, targetMethods, targetMethodKeys);
				} else {
					cs = new CallSite(method, internalMKey, pos, length,
							targetMethods, targetMethodKeys);
				}
				Set targets = rcs.resolvedTargets;
				for (Iterator it1 = targets.iterator(); it1.hasNext();) {
					IWorklistMethod tgt = (IWorklistMethod) it1.next();
					IMethod iMtgt = (IMethod) internalToExternalMethods
							.get(tgt);
					String tgtKey = tgt.getKey();
					targetMethods.add(iMtgt);
					targetMethodKeys.add(tgtKey);
				}
				return cs;
			}
		}
		return null;
	}

	public Set getCallingMethods(IMethod method) {
		if (!methodsPopulated) {
			populateMethods();
		}
		IWorklistMethod wmethod = (IWorklistMethod) externalToInternalMethods
				.get(method);
		Set result = new HashSet();
		if (wmethod == null) {
			return result;
		}
		for (Iterator it = reachableMethods.iterator(); it.hasNext();) {
			IWorklistMethod m = (IWorklistMethod) it.next();
			IMethod iM = (IMethod) internalToExternalMethods.get(m);
			if (iM != null) {
				Set resolvedCSs = m.getResolvedCallSites();
				for (Iterator it1 = resolvedCSs.iterator(); it1.hasNext();) {
					ResolvedCallSite rcs = (ResolvedCallSite) it1.next();
					if (rcs.methodName.equals(wmethod.getName())
							&& rcs.paramTypes.length == wmethod.getParameters().length) {
						Set targets = rcs.resolvedTargets;
						if (targets.contains(wmethod)) {
							//System.out.println(iM + " ---> \n" + method +
							// "\n");
							result.add(iM);
						}
					}
				}
			}
		}
		return result;
	}

	// private helpers
	private IType getIType(IWorklistType type) {
		String typeName = type.getCompleteName();
		IType iType = null;
		// search for the corresponding IType
		// if it is anonymous, search differently
		if (type instanceof ASTType && /*
									    * ((ASTType)
									    * type).getSimpleName().equals("")
									    */
		((ASTType) type).getPosition() > 0) {
			iType = getAnonymousType((ASTType) type);
			//System.out.println(typeName + " is anonymous " + (iType ==
			// null));
			return iType;
		}

		boolean found = false;
		for (Iterator it = javaProjects.iterator(); !found && it.hasNext();) {
			IJavaProject ijp = (IJavaProject) it.next();
			try {
				iType = ijp.findType(typeName);
				if (iType != null)
					found = true;
			} catch (JavaModelException jme) {
				System.out.println(jme.getMessage());
			}
		}
		// try a second search if not found
		if (iType == null) {
			iType = getSecondaryIType(type.getCompleteName(), type
					.getPackageName(), (type instanceof ASTType));
		}
		return iType;
	}

	// get the IType from the name of the type
	// when the type is secondary type
	// or otherwise not available by regular find
	private IType getSecondaryIType(String completeName, String packageName,
			boolean isFromAST) {
		String[] possTypeName = getAllPossSimpleNames(completeName, packageName);
		IType result = null;
		for (Iterator it = javaProjects.iterator(); result == null
				&& it.hasNext();) {
			IJavaProject ijProj = (IJavaProject) it.next();
			try {
				IPackageFragment[] chl = ijProj.getPackageFragments();
				IPackageFragment pkg = null;
				for (int i = 0; i < chl.length; i++) {
					if (chl[i].getElementName().equals(packageName)) {
						pkg = chl[i];
						break;
					}
				}
				if (pkg != null) {
					ICompilationUnit[] units = pkg.getCompilationUnits();
					for (int i = 0; i < units.length; i++) {
						IType[] types = units[i].getAllTypes();
						for (int j = 0; j < types.length; j++) {
							for (int k = 0; k < possTypeName.length; k++) {
								if (types[j].getTypeQualifiedName().equals(
										possTypeName[k])) {
									result = types[j];
								}
							}
						}
					}
				}
			} catch (JavaModelException jme) {
				jme.printStackTrace();
			}
		}
		if (isFromAST) {
			//System.out.println("AST TYPE:" + completeName + " " +
			// result.getFullyQualifiedName());
			result = getInnerDeclaredType(result, completeName);
			return result;
		} else {
			// weird
			System.out.println("IType not found for " + completeName);
		}
		return result;
	}

	// recursive method to find named types declared inside
	// other types or methods
	private IType getInnerDeclaredType(IJavaElement elem, String name) {
		if (elem == null)
			return null;

		if (elem instanceof IType
				&& ((IType) elem).getFullyQualifiedName().equals(name)) {
			return (IType) elem;
		} else if (elem instanceof IType) {
			try {
				IJavaElement[] children = ((IType) elem).getChildren();
				for (int i = 0; i < children.length; i++) {
					IType result = getInnerDeclaredType(children[i], name);
					if (result != null)
						return result;
				}
			} catch (JavaModelException jme) {
				System.out.println(jme.getMessage());
			}
		} else if (elem instanceof IMethod) {
			try {
				IJavaElement[] children = ((IMethod) elem).getChildren();
				for (int i = 0; i < children.length; i++) {
					IType result = getInnerDeclaredType(children[i], name);
					if (result != null)
						return result;
				}
			} catch (JavaModelException jme) {
				System.out.println(jme.getMessage());
			}
		}
		return null;
	}

	private IType getAnonymousType(ASTType type) {
		IType result = getSecondaryIType(type.getParentName(), type
				.getPackageName(), true);
		if (result != null) {
			result = getInnerDeclaredType(result, type.getPosition());
		} else {
			System.out.println("NOT FOUND ANON PARENT TYPE "
					+ type.getParentName() + " " + type.getPackageName());
		}
		return result;
	}

	// recursive method to find named types declared inside
	// other types or methods
	private IType getInnerDeclaredType(IJavaElement elem, int pos) {
		//System.out.println("getInnerDeclaredType:" + elem.getElementName());
		if (elem == null)
			return null;

		if (elem instanceof IType) {
			try {
				IType t = (IType) elem;
				if (t.getSourceRange().getOffset() == pos)
					return t;
				IJavaElement[] children = ((IType) elem).getChildren();
				for (int i = 0; i < children.length; i++) {
					IType result = getInnerDeclaredType(children[i], pos);
					if (result != null)
						return result;
				}
			} catch (JavaModelException jme) {
				System.out.println(jme.getMessage());
			}
		} else {
			IJavaElement[] children = null;
			try {
				if (elem instanceof IMethod) {
					children = ((IMethod) elem).getChildren();
				} else if (elem instanceof IField) {
					children = ((IField) elem).getChildren();
				} else if (elem instanceof IInitializer) {
					children = ((IInitializer) elem).getChildren();
				}
			} catch (JavaModelException jme) {
				System.out.println(jme.getMessage());
			}
			for (int i = 0; children != null && i < children.length; i++) {
				IType result = getInnerDeclaredType(children[i], pos);
				if (result != null)
					return result;
			}
		}
		return null;
	}

	// returns all possible simple names of this class type
	// to be used when the above method fails
	public String[] getAllPossSimpleNames(String completeName,
			String packageName) {
		String name = completeName;
		// remove package
		if (completeName.indexOf(packageName) == 0 && !packageName.equals("")) {
			name = completeName.substring(packageName.length() + 1);
		}
		String[] result = new String[name.split("\\$|\\.").length];
		for (int i = 0; i < result.length; i++) {
			if (i > 0) {
				result[i] = result[i - 1] + name.charAt(0);
				name = name.substring(1);
			} else {
				result[i] = "";
			}
			int index;
			int indexDot = name.indexOf(".");
			int indexDollar = name.indexOf("$");
			if (indexDot > -1)
				if (indexDollar > -1)
					index = (indexDot > indexDollar) ? indexDollar : indexDot;
				else
					index = indexDot;
			else if (indexDollar > -1)
				index = indexDollar;
			else
				index = name.length();
			result[i] += name.substring(0, index);
			name = name.substring(index);
		}
		return result;
	}

	// getting the IMethod for an existing IWorklistMethod
	private IMethod getIMethod(IWorklistMethod iwm) {
		IMethod result = null;
		IWorklistType declaringClass = iwm.getDeclaringClass();
		// search for the corresponding IType
		IType declaringIType = getIType(declaringClass);
		// given the IType, search for the IMethod

		if (declaringIType != null) {
			try {
				IMethod[] allMethods = declaringIType.getMethods();
				for (int i = 0; i < allMethods.length; i++) {
					IMethod crt = allMethods[i];
					if ((iwm.getName().equals(crt.getElementName()) || (iwm
							.getName().equals("<init>") && crt.isConstructor()))
							&& iwm.getParameters().length == crt
									.getNumberOfParameters()) {
						String[] params = crt.getParameterTypes();
						boolean found = true;
						for (int j = 0; found && j < params.length; j++) {
							String compareTo = iwm.getParameters()[j]
									.getCompleteName();
							//System.out.println("compareto:" + iwm);
							String compareToSign = Signature
									.createTypeSignature(compareTo, true);
							if (!params[j].equals(compareToSign)) {
								compareToSign = Signature.createTypeSignature(
										compareTo, false);
								if (!params[j].equals(compareToSign)) {
									compareToSign = Signature
											.createTypeSignature(Signature
													.getSimpleName(compareTo),
													false);
									if (!params[j].equals(compareToSign)) {
										if (iwm.getParameters()[j] instanceof ASTType) { // last
											// try
											String simpleName = ((ASTType) (iwm
													.getParameters()[j]))
													.getSimpleName();
											if (!simpleName.equals("")) {
												compareToSign = Signature
														.createTypeSignature(
																simpleName,
																false);
												if (!params[j]
														.equals(compareToSign)) {
													found = false;
												}
											} else {
												found = false;
											}
										} else {
											found = false;
										}
									}
								}
							}
						}
						if (found) {
							return crt;
						}
					}
				}
			} catch (JavaModelException jme) {
				System.out.println(jme.getMessage());
			}
		} else {
			System.out.println("Declaring class for " + iwm.getKey()
					+ " is null");
		}
		return result;
	}

	// fills the maps that relate equivalent method representations
	private void populateMethods() {
		methodsPopulated = true;
		for (Iterator it = reachableMethods.iterator(); it.hasNext();) {
			IWorklistMethod iwm = (IWorklistMethod) it.next();
			IMethod im = getIMethod(iwm);
			// test
			if (im == null)
				System.out.println("IMethod not found for " + iwm);
			else {
				internalToExternalMethods.put(iwm, im);
				externalToInternalMethods.put(im, iwm);
			}
		}
	}

}