/*
 * Created on May 19, 2005
 *
 */
package tacle.internal.rta;

import java.io.PrintWriter;
import java.util.Map;
import java.util.HashMap;
import java.util.Set;
import java.util.HashSet;
import java.util.Iterator;

import tacle.internal.rta.data.ASTType;
import tacle.internal.rta.data.InfoDictionary;
import tacle.internal.rta.data.IWorklistType;
import tacle.internal.types.ASTManager;
import tacle.internal.types.TacleTypesException;

import org.eclipse.jdt.core.IMethod;
import org.eclipse.jdt.core.JavaModelException;
import org.eclipse.jdt.core.dom.CompilationUnit;
//import org.eclipse.jdt.core.dom.IBinding;
import org.eclipse.jdt.core.ICompilationUnit;

//import org.eclipse.jdt.internal.corext.dom.Bindings;

/**
 * @author Mariana L Barca
 * 
 * Manages the type hierarchy which is built in the beginning by the type
 * analysis classes
 */
public class TypeManager {
	// holds direct subclasses
	private Map typeHierarchy; // IWorklistClass -> Set of IWorklistClass

	// holds all subclasses
	// it's a cache to avoid recursive calls many on the above structure
	private Map allTypeHierarchy; // IWorklistClass -> Set of IWorklistClass

	private CompilationUnit mainMethodCU;

	private ASTManager astMgr;

	// class that declares the startup method
	private IWorklistType mainDeclaringClass;

	// name of the startup method
	private String mainName;
	
	private int mainLocationInSource;
	
	// enables a cache for easy access to ASTs
	private final boolean keepASTs = true;

	// constructor
	public TypeManager(IMethod m) throws TacleTypesException {
		astMgr = new ASTManager(keepASTs);
		//TacleTypes.getInstance().analyze(m, astMgr, true);

		//typeHierarchy = TacleTypes.getInstance().getTypeHierarchy();
		//compilationUnits = TacleTypes.getInstance().getCompilationUnits();
		//mainMethodCU = astMgr.createCompilationUnit(m.getCompilationUnit());

		typeHierarchy = new HashMap();
		allTypeHierarchy = new HashMap();
		mainMethodCU = astMgr.createCompilationUnit(m.getCompilationUnit());
		String mainDeclPkg = m.getDeclaringType().getPackageFragment()
				.getElementName();
		String mainDeclCls = m.getDeclaringType().getFullyQualifiedName();
		mainDeclaringClass = InfoDictionary.v().getMainClass(mainDeclPkg,
				mainDeclCls);
		// just for test
		//displayTypeHierarchy(typeHierarchy);
		
		mainName = m.getElementName();
		try {
			// location in source
			mainLocationInSource = m.getSourceRange().getOffset();
			if (m.isConstructor()) {
				mainName = m.getElementName();
			}
		}catch(JavaModelException jme) {
			System.out.println(jme.getMessage());
		}

	}

	// uses ast manager to get compilation unit for a source file
	public CompilationUnit getCU(ICompilationUnit icu) {
		try {
			return astMgr.createCompilationUnit(icu);
		} catch (TacleTypesException tte) {
			tte.printStackTrace();
		}
		return null;
	}

	// compilation unit of the main method
	public CompilationUnit getMainMethodCU() {
		return mainMethodCU;
	}

	public IWorklistType getMainDeclaringClass() {
		return mainDeclaringClass;
	}

	public String getMainName() {
		return mainName;
	}
	
	public int getMainLocationInSource() {
		return mainLocationInSource;
	}
	
	// returns only the direct subtypes
	// to get all subtypes, call method below
	public Set getDirectSubtypes(IWorklistType type) {
		if (typeHierarchy.containsKey(type))
			return (Set) typeHierarchy.get(type);
		else
			return null;
	}

	/**
	 * returns set of strings or ClassInfos representing all subclasses calls
	 * recursive method findAllSubtypes
	 */
	public Set getAllSubtypes(IWorklistType type) {
		Set result = new HashSet();
		result.add(type);
		findAllSubtypes(type, result);
		allTypeHierarchy.put(type, result);
		return result;
	}

	// recursive
	private void findAllSubtypes(IWorklistType type, Set result) {
		if (typeHierarchy.containsKey(type)) {
			Set subclasses = (Set) typeHierarchy.get(type);
			result.addAll(subclasses);
			for (Iterator it = subclasses.iterator(); it.hasNext();) {
				findAllSubtypes((IWorklistType) it.next(), result);
			}
		}
	}

	// parent, child could be String or ClassInfo
	// adds pair to type hierarchy
	public void addInHierarchy(IWorklistType parent, IWorklistType child) {
		// test, keep
		if (child instanceof ASTType
				&& child.getCompleteName().equals("java.lang.String")) {
			throw new RuntimeException("STOP");
		}

		if (!typeHierarchy.containsKey(parent)
				|| typeHierarchy.get(parent) == null) {
			typeHierarchy.put(parent, new HashSet());
		}
		Set subCls = (Set) typeHierarchy.get(parent);
		subCls.add(child);
	}

	// used in testing, printsthe type hierarchy
	public void printHierarchy(PrintWriter pw) {
		for (Iterator it = typeHierarchy.keySet().iterator(); it.hasNext();) {
			IWorklistType key = (IWorklistType) it.next();
			if (pw == null) {
				System.out.println(key);
			} else {
				pw.println(key);
			}
			Set set = (Set) typeHierarchy.get(key);
			for (Iterator it1 = set.iterator(); it1.hasNext();) {
				IWorklistType subcls = (IWorklistType) it1.next();
				if (pw == null) {
					System.out.println(" -> " + subcls.getClass() + " "
							+ subcls);
				} else {
					pw.println(" -> " + subcls.getClass() + " " + subcls);
				}
			}
		}
	}
}
