/*
 * Created on Jun 24, 2005
 *
 */
package tacle.internal.rta.data;

import java.util.HashSet;
import java.util.Set;

import tacle.internal.rta.ResolvedCallSite;

/**
 * @author Mariana L Barca
 * 
 * Representation of methods from ASTs
 */
public class ASTMethod implements IWorklistMethod {

	private String key;

	private IWorklistType declaringClass;

	private String name;

	private IWorklistType returnType;

	private IWorklistType[] parameters;

	private Set resolvedCallSites;

	// constructor
	public ASTMethod(String k, String n, IWorklistType dC, IWorklistType rT,
			IWorklistType[] ps) {
		key = k;
		name = n;
		declaringClass = dC;
		returnType = rT;
		parameters = ps;
		resolvedCallSites = new HashSet();

		location = -1;
	}

	private int location;

	private ASTType parent;

	public ASTMethod(String n, ASTType p, int l) {
		name = n;
		parent = p;
		location = l;

		key = "";
		resolvedCallSites = new HashSet();
	}

	// getters
	public Set getResolvedCallSites() {
		return resolvedCallSites;
	}

	public IWorklistType getDeclaringClass() {
		return declaringClass;
	}

	public String getKey() {
		return key;
	}

	public String getName() {
		return name;
	}

	public IWorklistType[] getParameters() {
		return parameters;
	}

	public IWorklistType getReturnType() {
		return returnType;
	}

	public IWorklistType getParent() {
		return parent;
	}
	
	public int getLocation() {
		return location;
	}
	
	// setter
	public void addResolvedCallSite(ResolvedCallSite rcs) {
		resolvedCallSites.add(rcs);
	}

	public void setKey(String k) {
		key = k;
	}
	
	public void setDeclaringClass(IWorklistType dc) {
		declaringClass = dc;
	}
	
	public void setReturnType(IWorklistType rt) {
		returnType = rt;
	}
	
	public void setParameters(IWorklistType[] p) {
		parameters = p;
	}
	
	// overriding methods from class Object
	public String toString() {
		String pNames = "";
		for (int i = 0; i < getParameters().length - 1; i++) {
			pNames += getParameters()[i] + ",";
		}
		if (getParameters().length > 0) {
			pNames += getParameters()[getParameters().length - 1];
		}
		return getReturnType().toString() + " " + getName() + "(" + pNames
				+ ") KEY:" + getKey();
	}

	public boolean equals(Object other) {
		if (other instanceof ASTMethod) {
			ASTMethod astM = (ASTMethod) other;
			return this.key.equals(astM.key)
					|| ((this.parent != null) && astM.parent != null
							&& this.parent.equals(astM.parent)
							&& this.location != -1 && this.location == astM.location);
		}
		return false;

	}

	public int hashCode() {
		return this.name.hashCode();
	}
}
