/*
 * Created on Aug 28, 2005
 *
 */
package tacle.internal.rta.data;

import org.eclipse.jdt.core.dom.ITypeBinding;

/**
 * @author Mariana L Barca
 *  
 */
public class ASTType implements IWorklistType {
	private String completeName;

	private String simpleName;
	
	private String packageName;

	// this boolean is true for types other than
	// class types that appear in the library
	// in the "otherTypes" file (includes primitive types)
	// and false for types coming from user code
	private boolean fromSummary;

	// the following 2 fields are needed for identifying anonymouos
	// classes in the IType hierarchy
	private String declaringTypeBinaryName;

	private int anonTypePos;

	// constructors
	public ASTType(ITypeBinding itb, boolean fS) {
		
		completeName = itb.getQualifiedName();
		if (completeName.equals("")) {
			completeName = itb.getBinaryName();
		}

		// test this?
		if (!itb.isTopLevel() && !itb.isPrimitive() && !itb.isArray())
			completeName = itb.getBinaryName();

		if (itb.getPackage() == null)
			packageName = "";
		else {
			packageName = itb.getPackage().getName();
			if (packageName == null || packageName.equals("UNNAMED")) {
				packageName = "";
			}
		}
		fromSummary = fS;
		simpleName = itb.getName();
	}

	public ASTType(String pkg, String cls, String simpleName, boolean fS) {
		completeName = cls;
		packageName = pkg;
		fromSummary = fS;
		this.simpleName = simpleName;
	}

	public void setPosition(int pos, String parentName) {
		declaringTypeBinaryName = parentName;
		//System.out.println(completeName + " parent: " + declaringTypeBinaryName + " pos:" + pos);
		anonTypePos = pos;
	}
	
	public String getParentName() {
		return declaringTypeBinaryName;
	}
	
	public int getPosition() {
		return anonTypePos; 
	}
	
	// interface methods
	public String getCompleteName() {
		return completeName;
	}

	public String getPackageName() {
		return packageName;
	}

	public boolean isFromSummary() {
		return fromSummary;
	}

	// helpers

	public String getSimpleName() {
		return simpleName;
	}
	
	// returns the simple name of this class type
	// tricky with inner classes, may fail...
	//	public String getSimpleName() {
	//		String name = completeName;
	//		// remove package
	//		if (completeName.indexOf(packageName) == 0 && !packageName.equals("")) {
	//			name = completeName.substring(packageName.length() + 1);
	//		}
	//		// static inner class
	//		if (name.indexOf(".") >= 0)
	//			name = name.substring(0, name.indexOf("."));
	//		// inner class names
	//		if (name.indexOf("$") >= 0)
	//			name = name.substring(0, name.indexOf("$"));
	//		return name;
	//	}

	// returns all possible simple names of this class type
	// to be used when the above method fails
	public String[] getAllPossSimpleNames() {
		String name = completeName;
		// remove package
		if (completeName.indexOf(packageName) == 0 && !packageName.equals("")) {
			name = completeName.substring(packageName.length() + 1);
		}
		String[] result = new String[name.split("\\$|\\.").length];
		for (int i = 0; i < result.length; i++) {
			if (i > 0) {
				result[i] = result[i - 1] + name.charAt(0);
				name = name.substring(1);
			} else {
				result[i] = "";
			}
			int index;
			int indexDot = name.indexOf(".");
			int indexDollar = name.indexOf("$");
			if (indexDot > -1)
				if (indexDollar > -1)
					index = (indexDot > indexDollar) ? indexDollar : indexDot;
				else
					index = indexDot;
			else if (indexDollar > -1)
				index = indexDollar;
			else
				index = name.length();
			result[i] += name.substring(0, index);
			name = name.substring(index);
		}
		return result;
	}

	// overriding methods of class Object
	public int hashCode() {
		return completeName.hashCode();
	}

	public boolean equals(Object other) {
		if (other instanceof ASTType) {
			ASTType astC = (ASTType) other;
			if (packageName.equals(astC.packageName)
					&& completeName.equals(astC.completeName))
				return true;
		}
		return false;
	}

	public String toString() {
		return completeName;
	}
}
