/*
 * Created on Jun 22, 2005
 *
 */
package tacle.internal.rta.data;

/**
 * @author Mariana L Barca
 *  
 */
public class CallSiteInfo implements IWorklistCallSite {
	// enclosing method
	private int packageId;

	private int methodId;

	private MethodInfo method;

	// var type
	private int varClassId;

	private int varPackageId;

	private ClassInfo varClass;

	// static target method
	private int targetPackageId;

	private int targetMethodId;

	private MethodInfo targetMethod;

	// virtual info
	private boolean virtual;

	// methods that populate info
	private void populateEnclosingMethod() {
		method = InfoDictionary.v().getMethodInfoForId(methodId, packageId);
	}

	private void populateVarType() {
		varClass = InfoDictionary.v().getClassInfo(varClassId, varPackageId);
	}

	private void populateTargetMethod() {
		targetMethod = InfoDictionary.v().getMethodInfoForId(targetMethodId,
				targetPackageId);
	}

	// constructors
	public CallSiteInfo() {
	}

	// getters & setters
	public MethodInfo getMethod() {
		if (method == null)
			this.populateEnclosingMethod();
		return method;
	}

	public void setMethod(MethodInfo method) {
		this.method = method;
	}

	public int getMethodId() {
		return methodId;
	}

	public void setMethodId(int methodId) {
		this.methodId = methodId;
	}

	public int getPackageId() {
		return packageId;
	}

	public void setPackageId(int packageId) {
		this.packageId = packageId;
	}

	public MethodInfo getTargetMethod() {
		if (targetMethod == null)
			this.populateTargetMethod();
		return targetMethod;
	}

	public void setTargetMethod(MethodInfo targetMethod) {
		this.targetMethod = targetMethod;
	}

	public int getTargetMethodId() {
		return targetMethodId;
	}

	public void setTargetMethodId(int targetMethodId) {
		this.targetMethodId = targetMethodId;
	}

	public int getTargetPackageId() {
		return targetPackageId;
	}

	public void setTargetPackageId(int targetPackageId) {
		this.targetPackageId = targetPackageId;
	}

	public ClassInfo getVarClass() {
		if (varClass == null)
			this.populateVarType();
		return varClass;
	}

	public void setVarClass(ClassInfo varClass) {
		this.varClass = varClass;
	}

	public int getVarClassId() {
		return varClassId;
	}

	public void setVarClassId(int varClassId) {
		this.varClassId = varClassId;
	}

	public int getVarPackageId() {
		return varPackageId;
	}

	public void setVarPackageId(int varPackageId) {
		this.varPackageId = varPackageId;
	}

	public boolean isVirtual() {
		return virtual;
	}

	public void setVirtual(boolean virtual) {
		this.virtual = virtual;
	}

	// overriding methods from class Object
	public boolean equals(Object o) {
		if (o instanceof CallSiteInfo) {
			CallSiteInfo csi = (CallSiteInfo) o;
			return (this.methodId == csi.methodId
					&& this.targetMethodId == csi.targetMethodId && this.varClassId == csi.varClassId);
		} else
			return false;
	}

	public int hashCode() {
		return methodId + targetMethodId + varClassId;
	}
}
