/*
 * Created on Jun 24, 2005
 *
 */
package tacle.internal.rta.data;

import java.io.BufferedReader;
import java.io.FileReader;
import java.io.IOException;
import java.util.HashMap;
import java.util.StringTokenizer;
import java.util.Set;
import java.util.HashSet;
import java.util.Map;

import org.eclipse.core.runtime.Path;
import org.eclipse.core.runtime.Platform;
import org.osgi.framework.Bundle;

/**
 * @author Mariana L Barca
 * 
 * Singleton Accessed globally by all Info classes, as needed
 * reads info from the summary files
 */
public class DataReader {

	private String dataPath = "C:\\tacle_summary";

	private String fileSeparator;

	// private constructor
	private DataReader() {
		fileSeparator = System.getProperty("file.separator");
		
		// finds the location of the library summary
		// we assume it is in the installation directory of the eclipse plugin
		// named edu.ohio-state.cse.presto.tacle, in the subdirectory summary
		//Bundle b = Platform.getBundle("ca.mcgill.sable.soot");
		Bundle b = Platform.getBundle("edu.ohio-state.cse.presto.tacle");
		// in case we need the version, this is how to get it
		//String version = (String)b.getHeaders().get("Bundle-Version"));
		/*try {
			this.dataPath = Platform.asLocalURL(Platform.find(b, new Path("summary"))).getPath();
		} catch (IOException ioe) {
			System.out.println(ioe.getMessage());
			//ioe.printStackTrace();
			// if summary not found, plugin exits in a rather abrupt way, for now
			throw new RuntimeException("Exiting...");
		}*/
	}

	// unique instance
	private static DataReader instance = null;

	public static DataReader v() {
		if (instance == null) {
			instance = new DataReader();
		}
		return instance;
	}

	// divides str into token and returns an array
	// that contains them
	private String[] parseTokens(String str, String delim) {
		StringTokenizer st = new StringTokenizer(str, " ");
		int count = st.countTokens();
		String[] values = new String[count];
		int i = 0;
		while (st.hasMoreTokens()) {
			values[i] = st.nextToken();
			i++;
		}
		return values;
	}

	// loads the file packages 
	public void loadAllPackages(Map packageIdByName, Map packageNameById) {
		String packagesFile = dataPath + fileSeparator + "packages";
		try {
			BufferedReader in = new BufferedReader(new FileReader(packagesFile));
			String str = "";
			// format: packageid packagename
			while ((str = in.readLine()) != null) {
				String id = str.substring(0, str.indexOf(' '));
				String name = str.substring(str.indexOf(' ')+1);
				packageIdByName.put(name, Integer.valueOf(id));
				packageNameById.put(Integer.valueOf(id), name);
			}
			in.close();
		} catch (IOException ioe) {
			ioe.printStackTrace();
		}
	}

	public void loadAllMethods(Integer packageId, Map methodInfos,
			Map methodInfosByClassId) {
		String methodsFile = dataPath + fileSeparator + packageId
				+ fileSeparator + "methods";
		try {
			BufferedReader in = new BufferedReader(new FileReader(methodsFile));
			String str = "";
			while ((str = in.readLine()) != null) {
				// format:
				// methodid methodname declpkgid declclsid returnpkgid
				// returntypeid paramno parampkgid
				// paramtypeid... modifiers
				String[] values = parseTokens(str, " ");
				MethodInfo mi = new MethodInfo();
				mi.setMethodId(Integer.parseInt(values[0]));
				mi.setName(values[1]);
				mi.setDeclaringPackageId(Integer.parseInt(values[2]));
				mi.setDeclaringClassId(Integer.parseInt(values[3]));
				mi.setReturnPackageId(Integer.parseInt(values[4]));
				mi.setReturnTypeId(Integer.parseInt(values[5]));
				int[][] param = new int[Integer.parseInt(values[6])][2];
				for (int i = 0; i < param.length; i++) {
					param[i][0] = Integer.parseInt(values[7 + 2 * i]);
					param[i][1] = Integer.parseInt(values[8 + 2 * i]);
				}
				mi.setParamTypeIds(param);
				mi.setModifiers(Integer.parseInt(values[7 + 2 * param.length]));
				methodInfos.put(Integer.valueOf(values[0]), mi);

				// add to methodInfosByClassId: format:
				// Map methodInfosByClassId; //Integer->Map//String
				// methodName->Set of MethodInfos
				Integer classId = Integer.valueOf(values[3]);
				Map methodsByName = (Map) methodInfosByClassId.get(classId);
				if (methodsByName == null) {
					methodsByName = new HashMap();
					methodInfosByClassId.put(classId, methodsByName);
				}
				String methodName = values[1];
				Set methods = (Set) methodsByName.get(methodName);
				if (methods == null) {
					methods = new HashSet();
					methodsByName.put(methodName, methods);
				}
				methods.add(mi);
			}
			in.close();
		} catch (IOException ioe) {
			ioe.printStackTrace();
		}
	}

	// loads file classes from the specified package 
	public void loadAllClasses(Integer packageId, Map classInfos,
			Map classInfosByName) {
		String classesFile = dataPath + fileSeparator + packageId
				+ fileSeparator + "classes";
		try {
			BufferedReader in = new BufferedReader(new FileReader(classesFile));
			String str = "";
			while ((str = in.readLine()) != null) {
				// format:
				// classid name completename superclasspkgid superclassid
				// interfno
				// interfpkgid interfid... modifiers
				String[] values = parseTokens(str, " ");
				ClassInfo ci = new ClassInfo();
				ci.setClassId(Integer.parseInt(values[0]));
				ci.setPackageId(packageId.intValue());
				ci.setCompleteName(values[1]);
				ci.setSuperPackageId(Integer.parseInt(values[2]));
				ci.setSuperClassId(Integer.parseInt(values[3]));
				int[][] interf = new int[Integer.parseInt(values[4])][2];
				for (int i = 0; i < interf.length; i++) {
					interf[i][0] = Integer.parseInt(values[5 + 2 * i]);
					interf[i][1] = Integer.parseInt(values[6 + 2 * i]);
				}
				ci.setInterfaceIds(interf);
				ci.setModifiers(Integer.parseInt(values[2 * interf.length + 5]));
				classInfos.put(Integer.valueOf(values[0]), ci);
				classInfosByName.put(ci.getCompleteName(), ci);
			}
			in.close();
		} catch (IOException ioe) {
			ioe.printStackTrace();
		}
	}

	// loads global file otherTypes
	public void loadAllOtherTypes(Map types, Map typesByName) {
		String typesFile = dataPath + fileSeparator + "otherTypes";
		try {
			BufferedReader in = new BufferedReader(new FileReader(typesFile));
			String str = "";
			while ((str = in.readLine()) != null) {
				// format:
				// typeid typename 
				String[] values = parseTokens(str, " ");
				ASTType cls = new ASTType("", values[1], "", true);
				types.put(Integer.valueOf(values[0]), cls);
				typesByName.put(values[1], cls);
			}
			in.close();
		} catch (IOException ioe) {
			ioe.printStackTrace();
		}
	}

	// loads file methodInstantiatesClass from the specified package
	public void loadAllMethodInstantiatesClass(Integer packageId,
			Map methodsInfos) {
		String mthdInstClsFile = dataPath + fileSeparator + packageId
				+ fileSeparator + "methodInstantiatesClass";
		try {
			BufferedReader in = new BufferedReader(new FileReader(
					mthdInstClsFile));
			String str = "";
			while ((str = in.readLine()) != null) {
				// methodid packageid classid
				String[] values = parseTokens(str, " ");
				Integer methodId = Integer.valueOf(values[0]);
				Integer classPkgId = Integer.valueOf(values[1]);
				Integer classId = Integer.valueOf(values[2]);
				MethodInfo mi = (MethodInfo) methodsInfos.get(methodId);
				mi.setInstantiatedClassId(classPkgId, classId);
			}
			in.close();
		} catch (IOException ioe) {
			ioe.printStackTrace();
		}
	}

	// loads file methodInitializesClass from the specified package
	public void loadAllMethodInitializesClass(Integer packageId,
			Map methodsInfos) {
		String mthdInstClsFile = dataPath + fileSeparator + packageId
				+ fileSeparator + "methodInitializesClass";
		try {
			BufferedReader in = new BufferedReader(new FileReader(
					mthdInstClsFile));
			String str = "";
			while ((str = in.readLine()) != null) {
				// methodid packageid classid
				String[] values = parseTokens(str, " ");
				Integer methodId = Integer.valueOf(values[0]);
				Integer classPkgId = Integer.valueOf(values[1]);
				Integer classId = Integer.valueOf(values[2]);
				MethodInfo mi = (MethodInfo) methodsInfos.get(methodId);
				mi.setInitializedClassId(classPkgId, classId);
			}
			in.close();
		} catch (IOException ioe) {
			ioe.printStackTrace();
		}
	}

	// loads file methodHasCallSite from the specified package
	// format of Map methodInfos: // Integer -> MethodInfo
	public void loadAllMethodHasCallSite(Integer packageId, Map methodInfos) {
		String mthdHasCallSiteFile = dataPath + fileSeparator + packageId
				+ fileSeparator + "methodHasCallSite";
		try {
			BufferedReader in = new BufferedReader(new FileReader(
					mthdHasCallSiteFile));
			String str = "";
			while ((str = in.readLine()) != null) {
				// methodid packageid classid targetpackageid targetmethodid
				// virtual(1)
				String[] values = parseTokens(str, " ");
				CallSiteInfo csi = new CallSiteInfo();
				int methodId = Integer.parseInt(values[0]);

				MethodInfo mi = InfoDictionary.v().getMethodInfoForId(methodId,
						packageId.intValue());

				csi.setPackageId(packageId.intValue());
				csi.setMethodId(methodId);

				int classPkgId = Integer.parseInt(values[1]);
				int classId = Integer.parseInt(values[2]);
				csi.setVarPackageId(classPkgId);
				csi.setVarClassId(classId);

				int targetPackageId = Integer.parseInt(values[3]);
				int targetMethodId = Integer.parseInt(values[4]);
				csi.setTargetPackageId(targetPackageId);
				csi.setTargetMethodId(targetMethodId);

				csi.setVirtual(values[5].equals("1"));
				
				mi.setCallSite(csi);
			}
			in.close();
		} catch (IOException ioe) {
			ioe.printStackTrace();
		}
	}
}

