/*
 * Created on May 16, 2005
 *
 */
package tacle.internal.types;

import java.util.HashMap;
import java.util.Map;

import org.eclipse.jdt.core.IClassFile;
import org.eclipse.jdt.core.ICompilationUnit;
import org.eclipse.jdt.core.IMember;
import org.eclipse.jdt.core.JavaModelException;
import org.eclipse.jdt.core.dom.AST;
import org.eclipse.jdt.core.dom.ASTParser;
import org.eclipse.jdt.core.dom.CompilationUnit;

/**
 * @author Mariana L Barca
 *
 * TACLE implementation of IASTManager  
 */
public class ASTManager implements IASTManager {

	private final boolean resolveBindings = true;

	// map of compilation units, used as a cache to avoid
	// the time spen tin parsing sources
	// member: String handle identifier of source file (usually class name) 
	// -> CompilationUnit
	private Map cUnits;

	private boolean keepUnits;

	// constructor
	public ASTManager(boolean keepUnits) {
		this.keepUnits = keepUnits;
		if (keepUnits)
			cUnits = new HashMap();
	}

	// private, so it cannot be called
	private ASTManager() {
	}
	
	// dispatcher method, further calls the appropriate methods 
	// unit can be ICompilationUnit or IClassFile
	public CompilationUnit createCompilationUnit(Object unit)
			throws TacleTypesException {
		if (unit instanceof IClassFile) {
			IClassFile classFile = (IClassFile) unit;
			return createCompilationUnit(classFile);
		} else if (unit instanceof ICompilationUnit) {
			ICompilationUnit icu = (ICompilationUnit) unit;
			return createCompilationUnit(icu);
		} else
			return null;
	}

	// parses the source file with an AST parser and returns the 
	// resulted CompilationUnit
	private CompilationUnit createCompilationUnit(ICompilationUnit icu)
			throws TacleTypesException {
		if (icu != null && icu.exists()) {
			CompilationUnit cu = null;
			if (keepUnits) {
				cu = (CompilationUnit) cUnits.get(icu.getHandleIdentifier());
				if (cu != null)
					return cu;
			}
			ASTParser parser = ASTParser.newParser(AST.JLS2);
			parser.setSource(icu);
			parser.setResolveBindings(resolveBindings);
			cu = (CompilationUnit) parser.createAST(null);
			if (keepUnits && cu != null) {
				cUnits.put(icu.getHandleIdentifier(), cu);
			}
			return cu;
		}

		return null;
	}

	// parses the source file associated to the class file
	// with an AST parser and returns the 
	// resulted CompilationUnit
	private CompilationUnit createCompilationUnit(IClassFile classFile)
			throws TacleTypesException {
		try {
			if (classFile != null && classFile.exists()
					&& classFile.getSource() != null) {
				CompilationUnit cu = null;
				if (keepUnits) {
					cu = (CompilationUnit) cUnits.get(classFile
							.getHandleIdentifier());
					if (cu != null)
						return cu;
				}
				ASTParser parser = ASTParser.newParser(AST.JLS2);
				parser.setSource(classFile);
				parser.setResolveBindings(resolveBindings);
				cu = (CompilationUnit) parser.createAST(null);
				if (keepUnits && cu != null) {
					cUnits.put(classFile.getHandleIdentifier(), cu);
				}
				return cu;
			} 
		} catch (JavaModelException e) {
			e.printStackTrace();
			throw new TacleTypesException(e.getMessage());
		}

		return null;
	}

	public Object getUnit(IMember member) {
		if (member.isBinary()) {
			IClassFile classFile = member.getClassFile();
			return classFile;
		} else {
			ICompilationUnit icu = member.getCompilationUnit();
			return icu;
		}
	}

}
